/*  Screem:  screem-tree-view.c
 *
 *  The tree view widget
 *
 *  Copyright (C) 2002  David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>

#include <string.h>

#include <libgnome/gnome-i18n.h>

#include <gtk/gtktreestore.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrenderertoggle.h>
#include <gtk/gtkcellrendererpixbuf.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkimage.h>
#include <gtk/gtktextbuffer.h>  /* needed to get feature modes
				   for the current document */
#include <gtk/gtkstock.h>

#include "screem-tree-view.h"

#include "screem-window.h"

#include "screem-markup.h"
#include "screem-dtd.h"

#include "libegg/dock/egg-dock.h"

#include "screemmarshal.h"

enum {
	TREE_VIEW_ICON_COL = 0,
	TREE_VIEW_NAME_COL,
	TREE_VIEW_TEXT_COL,
	TREE_VIEW_START_COL,
	TREE_VIEW_END_COL,
	TREE_VIEW_VALID_COL,
	TREE_VIEW_MAX_COL
};

enum {
	SELECT_CONTEXT,
	LAST_SIGNAL
};

static guint screem_tree_view_signals[ LAST_SIGNAL ] = { 0 };


static void screem_tree_view_class_init( ScreemTreeViewClass *klass );
static void screem_tree_view_init( ScreemTreeView *tree_view );
static void screem_tree_view_finalize( GObject *tree_view );
static void screem_tree_view_window_set( ScreemView *view );
static void screem_tree_view_size_request( GtkWidget *widget, 
					   GtkRequisition *req );
static void screem_tree_view_realize( GtkWidget *widget );
static void screem_tree_view_show( GtkWidget *widget );
static void screem_tree_view_hide( GtkWidget *widget );

static void screem_tree_view_display( ScreemView *view );
static void screem_tree_view_page( ScreemTreeView *view, ScreemPage *page );

static gboolean select_context( ScreemTreeView *view,
				GtkTreeModel *model, GtkTreeIter **it,
				guint pos, guint *start, guint *end,
				gboolean select_text,
				gboolean build );
static void page_change( ScreemPage *page, GParamSpec *spec, 
			 gpointer data );
static gboolean build_tree( ScreemTreeView *view );

static gboolean expand_trees( ScreemTreeView *view );

static gboolean screem_tree_view_clear_tree( GtkTreeModel *model, 
					     GtkTreePath *path, 
					     GtkTreeIter *iter, 
					     gpointer data );
static void screem_tree_view_build_tree( ScreemTreeView *view, 
					 const gchar *text );

static void screem_tree_view_create_mini_view( ScreemTreeView *view );
static void screem_tree_view_mini_activate( GtkTreeView *mini,
					    GtkTreePath *path,
					    GtkTreeViewColumn *column,
					    gpointer data );

struct ScreemTreeViewPrivate {
	GtkTreeModel *model;
	GtkTreeView *view;

	GtkTreeView *mini;

	gint timeout;
	ScreemPage *current;

	GtkTreeIter *context;
};

ScreemTreeView *screem_tree_view_new( ScreemWindow *window )
{
	ScreemTreeView *view;
	GType type;

	type = screem_tree_view_get_type();

	view = SCREEM_TREE_VIEW( g_object_new( type, "window", window, NULL) );

	return view;
}

gboolean screem_tree_view_select_element_context( ScreemTreeView *view,
						  guint pos,
						  guint *start, guint *end,
						  gboolean select_text )
{
	GtkTreeModel *model;
	guint tstart;
	guint tend;
	gboolean ret;

	g_return_val_if_fail( SCREEM_IS_TREE_VIEW( view ), FALSE );

	model = GTK_TREE_MODEL( view->private->model );

	if( ! start ) {
		start = &tstart;
	}
	if( ! end ) {
		end = &tend;
	}

	ret = select_context( view, 
			      model, NULL, pos, start, end, select_text, TRUE );
	if( ret ) {
		/* emit signal */
		g_signal_emit( G_OBJECT( view ),
			       screem_tree_view_signals[ SELECT_CONTEXT ],
			       0, *start, (*end) );
	}

	return ret;
}

gchar *screem_tree_view_query_element_context( ScreemTreeView *view,
					       guint pos,
					       gboolean query_text,
					       gboolean build,
					       guint *depth,
					       guint *start,
					       guint *end )
{
	GtkTreeModel *model;
	guint s;
	guint e;
	gboolean ret;
	gchar *retval;

	g_return_val_if_fail( SCREEM_IS_TREE_VIEW( view ), FALSE );

	model = GTK_TREE_MODEL( view->private->model );

	if( ! start ) {
		start = &s;
	}
	if( ! end ) {
		end = &e;
	}

	ret = select_context( view, 
			      model, NULL, pos, start, end,
			      query_text, build );
	retval = NULL;
	/* FIXME: view->private->context should always be non NULL
	   if ret is TRUE, however we don't handle <!DOCTYPE ... [ .. ]>
	   very well, and can cause it to be NULL, at least I think
	   that is what is causing a crash */
	if( ret && view->private->context ) {
		GValue value = { 0 };
		GtkTreePath *path;

		g_assert( view->private->context );

		gtk_tree_model_get_value( model, view->private->context,
					  TREE_VIEW_NAME_COL, &value );
		retval = (gchar*)g_value_get_string( &value );
		if( retval ) {
			retval = g_strdup( retval );
		}
		g_value_unset( &value );

		path = gtk_tree_model_get_path( model, view->private->context );

		if( depth ) {
			/* -1 as we are always under the [document] node */
			*depth = gtk_tree_path_get_depth( path );
			if( *depth > 0 ) {
				(*depth) --;
			}
		}
		gtk_tree_path_free( path );

		gtk_tree_iter_free( view->private->context );
		view->private->context = NULL;

	} else if( depth ) {
		*depth = 0;
	}

	return retval;
}

/* static stuff */
static void screem_tree_view_display( ScreemView *view )
{
	ScreemPage *page;

	g_object_get( G_OBJECT( view ), "page", &page, NULL );

	screem_tree_view_page( SCREEM_TREE_VIEW( view ), page );
}

static void screem_tree_view_page( ScreemTreeView *view, ScreemPage *page )
{
	if( view->private->current == page ) {
		return;
	}
	
	if( view->private->current ) {
		g_signal_handlers_disconnect_matched( G_OBJECT( view->private->current ),
						      G_SIGNAL_MATCH_DATA,
						      0, 0, NULL, NULL, view );
	}
	
	view->private->current = page;
	if( page ) {
		page_change( page, NULL, view );
		g_signal_connect( G_OBJECT( page ), "notify::changed",
				  G_CALLBACK( page_change ), view );
	}
	gtk_tree_view_expand_all( GTK_TREE_VIEW( view->private->view ) );
}


static gboolean select_context( ScreemTreeView *view,
				GtkTreeModel *model, GtkTreeIter **it,
				guint pos,
				guint *start, guint *end,
				gboolean select_text,
				gboolean build )
{
	gboolean loop;
	GtkTreeIter *tmp;
	GtkTreeIter root;
	
	/* If the build timer is going, remove it and construct a new tree,
	   but only if a build was requested */
	if( build && view->private->timeout != -1 ) {
		g_source_remove( view->private->timeout );
		view->private->timeout = -1;
		gdk_threads_leave();
		build_tree( view );
		gdk_threads_enter();
	}

	if( ! it ) {
		if( view->private->context ) {
			gtk_tree_iter_free( view->private->context );
			view->private->context = NULL;
		}

		loop = gtk_tree_model_get_iter_first( model, &root );
		if( loop && gtk_tree_model_iter_has_child( model, &root ) ) {
			GtkTreeIter *parent;
			
			parent = gtk_tree_iter_copy( &root );

			loop = gtk_tree_model_iter_children( model, &root, 
							     parent );
			if( loop ) {
				tmp = &root;
				it = &tmp;
			}
			gtk_tree_iter_free( parent );
		} else {
			loop = FALSE;
		}
	} else {
		loop = TRUE;
	}

	while( loop ) {
		GValue value = { 0 };
		gint estart;
		gint eend;

		gtk_tree_model_get_value( model, *it, TREE_VIEW_START_COL,
					  &value );
		estart = g_value_get_uint( &value );
		g_value_unset( &value );
		gtk_tree_model_get_value( model, *it, TREE_VIEW_END_COL,
					  &value );
		eend = g_value_get_uint( &value );
		g_value_unset( &value );

		if( pos <= eend && pos >= estart ) {
			/* in here somewhere */
			gboolean istext;
			
			gtk_tree_model_get_value( model, *it,
						  TREE_VIEW_TEXT_COL,
						  &value );
			istext = ( g_value_get_string( &value ) != NULL );
			g_value_unset( &value );

			if( ! istext || select_text ) {
				*start = estart;
				*end = eend + 1;
				if( view->private->context ) {
					gtk_tree_iter_free( view->private->context );
				}
				view->private->context = 
					gtk_tree_iter_copy( *it );
			}
			
			if( gtk_tree_model_iter_has_child( model, *it ) ) {
				GtkTreeIter child;
				
				if( gtk_tree_model_iter_children( model,
								  &child,
								  *it ) ) {
					GtkTreeIter *cptr;
					
					cptr = &child;
					select_context( view, 
							model, &cptr, pos,
							start, end,
							select_text,
							FALSE );
				}
			}
			break;
		} else {
			loop = gtk_tree_model_iter_next( model, *it );
		}
	}

	return loop;
}

static void page_change( ScreemPage *page, GParamSpec *spec, 
			 gpointer data )
{
	ScreemTreeView *view;
	gboolean status;

	view = SCREEM_TREE_VIEW( data );
	status = screem_page_get_changed( page );
	if( spec == NULL ) {
		status = TRUE;
	}
	
	if( view->private->timeout != -1 ) {
		g_source_remove( view->private->timeout );
	}

	view->private->timeout = g_timeout_add( 1000, 
						(GSourceFunc)build_tree,
						view );
}

static gboolean build_tree( ScreemTreeView *view )
{
	ScreemPage *page;
	GtkTreeModel *model;

	page = view->private->current;

	gdk_threads_enter();
	
	model = GTK_TREE_MODEL( view->private->model );

	gtk_tree_model_foreach( model, screem_tree_view_clear_tree, NULL );
	gtk_tree_store_clear( GTK_TREE_STORE( model ) );

	if( page ) {
		gchar *text;
		GtkTextBuffer *buffer;
		GtkTextTagTable *table;
		gboolean feature_markup;
		
		buffer = screem_page_get_text_buffer( page );
		table = gtk_text_buffer_get_tag_table( buffer );
		feature_markup = screem_page_is_markup( page );

		if( feature_markup ) {
			text = screem_page_get_data( page );
			screem_tree_view_build_tree( view, text );
			g_free( text );

			/* add with high priority so we get called before
			 * drawing funcs for the tree views */
			g_idle_add_full( 50, (GSourceFunc)expand_trees, view, NULL );
		}
	}

	gdk_threads_leave();
	
	return FALSE;
}

static gboolean expand_trees( ScreemTreeView *view )
{
	gtk_tree_view_expand_all( GTK_TREE_VIEW( view->private->mini ) );
	if( GTK_WIDGET_VISIBLE( view->private->view ) ) {
		gtk_tree_view_expand_all( GTK_TREE_VIEW( view->private->view ) );
	}

	return FALSE;
}

static gboolean screem_tree_view_clear_tree( GtkTreeModel *model, 
					     GtkTreePath *path,
					     GtkTreeIter *iter, 
					     gpointer data )
{

	return FALSE;
}

static void screem_tree_view_build_tree( ScreemTreeView *view, 
					 const gchar *text )
{
	gint pos;
	gint ppos;

	gint last;
	GSList *stack;

	GtkTreeModel *model;
	GtkTreeIter *parent;
	GtkTreeIter document;

	ScreemWindow *window;
	ScreemDTD *dtd;
	ScreemDTDClose pclose;
	const gchar *pname = NULL;

	gboolean gotroot;
	gboolean iscomment;

	gchar *tag;
	gint start;
	gint end;
	gchar *name;
	GtkTreeIter it;

	g_return_if_fail( SCREEM_IS_TREE_VIEW( view ) );
	g_return_if_fail( text != NULL );

	model = GTK_TREE_MODEL( view->private->model );

	g_object_get( G_OBJECT( view ), "window", &window, NULL );
	dtd = screem_window_get_current_dtd( window );
	last = strlen( text );
	stack = NULL;

	gtk_tree_store_append( GTK_TREE_STORE( model ),
				&document, NULL );
	gtk_tree_store_set( GTK_TREE_STORE( model ), 
			    &document,
			    TREE_VIEW_NAME_COL,
			    "[Document]",
			    TREE_VIEW_VALID_COL, TRUE,
			    -1 );
	parent = &document;

	pclose = MUST;
	gotroot = FALSE;

	for( ppos = pos = 0; text[ pos ]; ppos = pos ) {
		GValue value = { 0 };

		tag = screem_markup_next_tag( text, pos, 
						&start, &end, &name );
		
		if( ! tag ) {
			pos = last;
		} else {
			pos = start;
		}

		if( pos > ppos ) {
			/* add text node */
			GtkTreeIter tit;
			gchar *txt;
			gboolean valid;

			txt = g_strndup( text + ppos, pos - ppos );
			
			/* chomp / chop any white space, if all 
			   white space then we won't even bother
			   putting it in the tree */
			txt = g_strstrip( txt );

			if( *txt != '\0' ) {
				valid = screem_dtd_valid_child( dtd, pname, "#PCDATA" );
				if( ! valid ) {
					valid=screem_dtd_valid_child(dtd,
								     pname,
								     "CDATA");
				}

				gtk_tree_store_append( GTK_TREE_STORE( model ),
						       &tit, parent );
				gtk_tree_store_set( GTK_TREE_STORE( model ), 
						    &tit,
						    TREE_VIEW_NAME_COL,
						    "[text]",
						    TREE_VIEW_TEXT_COL,
						    txt,
						    TREE_VIEW_START_COL,
						    ppos,
						    TREE_VIEW_END_COL,
						    pos,
						    TREE_VIEW_VALID_COL,
						    valid,
						    -1 );
			}
		}

		if( tag ) {
			gchar firstc;
			pos = end + 1;
			
			
			if( ( iscomment = ! strcmp( "!--", name ) ) ) {
				g_free( name );
				name = g_strdup( "[Comment]" );
			}
		       
			firstc = *name;

			if( firstc == '/' ) {
				/* we are dealing with a closing tag */
				gboolean match;

				/* pop stack, change parent until
				   we hit the correct close tag or
				   we run out of stack */
				match = FALSE;
				while( stack && ! match ) {
					gtk_tree_store_set(GTK_TREE_STORE(model),
							   parent,
							   TREE_VIEW_END_COL, 
							   pos - 1, -1 );

					stack = g_slist_remove( stack, 
								stack->data );
					/* get parent name */
					gtk_tree_model_get_value( model, parent,
								  TREE_VIEW_NAME_COL,
								  &value );
					pname = g_value_get_string( &value );
					pclose = screem_dtd_element_close_state( dtd,
										 pname );
					match = ! strcmp( name + 1, pname );
					g_value_unset( &value );
					gtk_tree_iter_free( parent );
					if( stack ) {
						parent = stack->data;
					}
				}

				pname = NULL;
				if( ! stack ) {
					parent = &document;
				} 
				if( parent != &document ) {
					gtk_tree_model_get_value( model,
								  parent,
								  TREE_VIEW_NAME_COL,
								  &value );
					pname = g_value_get_string( &value );
					pclose = screem_dtd_element_close_state( dtd,
										 pname );
				}
			} else {
				gboolean valid;
				ScreemDTDClose close_state;

				close_state = 
					screem_dtd_element_close_state( dtd,
								       name );
				
			

				if( iscomment ) {
					valid = TRUE;
				} else if( firstc == '?' ||
					   !strcmp( "!DOCTYPE", name ) ) {
					valid = TRUE;
				} else if( parent == &document &&
					   ! gotroot &&
					   screem_dtd_is_root_element(dtd, 
								      name )) {
					valid = TRUE;
					gotroot = TRUE;
				} else {
					valid = screem_dtd_valid_child(dtd,
								       pname,
								       name);
				}
			
				/* if closing the parent tag is optional, and 
				   valid is FALSE, pop the parent */
				while( parent != &document && pclose == SHOULD &&
				       ! valid ) {
					/* assume this closes the parent */
					gtk_tree_store_set( GTK_TREE_STORE(model),
							    parent,
							    TREE_VIEW_END_COL, 
							    start, -1 );
					stack = g_slist_remove( stack, 
								stack->data );
					gtk_tree_iter_free( parent );
					if( stack ) {
						parent = stack->data;
					
						gtk_tree_model_get_value( model, parent,
									  TREE_VIEW_NAME_COL,
									  &value );
						pname = g_value_get_string( &value );
						pclose = screem_dtd_element_close_state( dtd,
											 pname );
						valid = screem_dtd_valid_child(dtd,
									       pname,
									       name);
						g_value_unset( &value );
					} else {
						parent = &document;
						pname = NULL;
					}
				}
			
				/* we are dealing with an opening tag */
				gtk_tree_store_append( GTK_TREE_STORE(model),
						       &it, parent );
				if( iscomment ) {
					gtk_tree_store_set(GTK_TREE_STORE( model ),
							   &it,
							   TREE_VIEW_TEXT_COL,
							   tag,
							   -1 );
				}


				/* set values for the iter */
				gtk_tree_store_set( GTK_TREE_STORE( model ),
						    &it,
						    TREE_VIEW_NAME_COL,
						    name,
						    TREE_VIEW_START_COL,
						    start,
						    TREE_VIEW_END_COL,
						    end,
						    TREE_VIEW_VALID_COL,
						    valid,
						    -1 );

				/* put it on the stack as a parent if needed,
				   we don't need to for:
				   1) comments / doctypes
				   2) tags closing themselves, eg <meta /> 
				   3) HTML tags which forbid closing, same as
				      2 but we need to lookup in the doctype */
				if( ( ! iscomment ) && 
				    ( firstc != '!' ) && ( firstc != '?' ) &&
				    ( text[ end - 1 ] != '/' ) &&
				    ( close_state != MUSTNT ) ) {
					parent = gtk_tree_iter_copy( &it );
					stack = g_slist_prepend( stack, 
								 parent );
					gtk_tree_model_get_value( model,
								  parent,
								  TREE_VIEW_NAME_COL,
								  &value );
					pname = g_value_get_string( &value );
					pclose = screem_dtd_element_close_state( dtd,
										 pname );
				}
				g_free( name );
				g_free( tag );
			}
		}
	}
	/* cleanup the stack */
	while( stack ) {
		gtk_tree_store_set(GTK_TREE_STORE(model),
				   parent,
				   TREE_VIEW_END_COL, 
				   pos, -1 );
		stack = g_slist_remove( stack, 
					stack->data );
		gtk_tree_iter_free( parent );
		if( stack ) {
			parent = stack->data;
		}
	}
}

static void screem_tree_view_create_mini_view( ScreemTreeView *view )
{
	ScreemTreeViewPrivate *private;
	GtkWidget *tree;
	GtkWidget *sw;
	GtkCellRenderer *rend;
	GtkCellRenderer *prend;
	GtkTreeViewColumn *col;
	ScreemWindow *window;
	GtkWidget *widget;
	
	private = view->private;

	tree = gtk_tree_view_new_with_model( GTK_TREE_MODEL(private->model) );

	private->mini = GTK_TREE_VIEW( tree );
	gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( tree ), TRUE );

	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
					GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );
	prend = gtk_cell_renderer_pixbuf_new();
	rend = gtk_cell_renderer_text_new ();
	col = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title( col, "Element" );
	gtk_tree_view_column_pack_start( col, prend, FALSE );
	gtk_tree_view_column_pack_start( col, rend, TRUE );
	gtk_tree_view_column_set_attributes( col, rend,
					     "text", TREE_VIEW_NAME_COL,
					     NULL );
	gtk_tree_view_column_set_attributes( col, prend,
					     "pixbuf",
					     TREE_VIEW_ICON_COL,
					     NULL );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), col );

	rend = gtk_cell_renderer_toggle_new();
	col = gtk_tree_view_column_new_with_attributes( "Valid",
							rend,
							"active",
							TREE_VIEW_VALID_COL,
							NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), col );
	
	gtk_container_add( GTK_CONTAINER( sw ), tree );
	gtk_widget_show( GTK_WIDGET( tree ) );
	gtk_widget_show( sw );

	g_object_get( G_OBJECT( view ), "window", &window, NULL );
	widget = egg_dock_item_new_with_stock( "Structure", _( "Tree" ), GTK_STOCK_INDEX,
					EGG_DOCK_ITEM_BEH_NORMAL );
	gtk_container_add( GTK_CONTAINER( widget ), sw );
	egg_dock_add_item( EGG_DOCK( window->dock ), EGG_DOCK_ITEM( widget ),
				EGG_DOCK_RIGHT );
	gtk_widget_show( widget );

	g_signal_connect( G_OBJECT( tree ), "row_activated",
			  G_CALLBACK( screem_tree_view_mini_activate ),
			  view );
}

static void screem_tree_view_mini_activate( GtkTreeView *mini, 
					    GtkTreePath *path,
					    GtkTreeViewColumn *column,
					    gpointer data )
{
	ScreemTreeView *view;
	GtkTreeModel *model;
	GtkTreeIter it;
	GValue value = { 0 };
	guint start;
	guint end;

	view = SCREEM_TREE_VIEW( data );

	model = gtk_tree_view_get_model( mini );
	gtk_tree_model_get_iter( GTK_TREE_MODEL( model ), &it, path );

	gtk_tree_model_get_value( model, &it, TREE_VIEW_START_COL,
				  &value );
	start = g_value_get_uint( &value );
	g_value_unset( &value );
	gtk_tree_model_get_value( model, &it, TREE_VIEW_END_COL,
				  &value );
	end = g_value_get_uint( &value );
	g_value_unset( &value );

	/* emit signal */
	if( start != end ) {
		g_signal_emit( G_OBJECT( view ),
			       screem_tree_view_signals[ SELECT_CONTEXT ],
			       0, start, end  + 1 );
	}
}

/* G Object stuff */
#define PARENT_TYPE SCREEM_TYPE_VIEW

static gpointer parent_class;

static void screem_tree_view_class_init( ScreemTreeViewClass *klass )
{
	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = G_OBJECT_CLASS( klass );
	widget_class = (GtkWidgetClass *)klass;
	parent_class = g_type_class_peek_parent( klass );

	object_class->finalize = screem_tree_view_finalize;

	widget_class->realize = screem_tree_view_realize;
	widget_class->size_request = screem_tree_view_size_request;
	widget_class->show = screem_tree_view_show;
	widget_class->hide = screem_tree_view_hide;

	screem_tree_view_signals[ SELECT_CONTEXT ] = 
		g_signal_new( "select_context",
			      G_OBJECT_CLASS_TYPE( object_class ),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET( ScreemTreeViewClass, 
					       select_context),
			      NULL, NULL,
			      screem_marshal_VOID__UINT_UINT,
			      G_TYPE_NONE, 2,
			      G_TYPE_UINT,
			      G_TYPE_UINT );
}

static void screem_tree_view_init( ScreemTreeView *tree_view )
{
	ScreemTreeViewPrivate *private;
	GtkWidget *sw;
	GtkCellRenderer *rend;
	GtkCellRenderer *prend;

	GtkTreeViewColumn *col;

	SCREEM_VIEW( tree_view )->window_set = screem_tree_view_window_set;
	SCREEM_VIEW( tree_view )->display = screem_tree_view_display;

	private = tree_view->private = g_new0( ScreemTreeViewPrivate, 1 );

	sw = gtk_scrolled_window_new( NULL, NULL );
	gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( sw ),
					GTK_POLICY_AUTOMATIC,
                                        GTK_POLICY_AUTOMATIC );
	gtk_scrolled_window_set_shadow_type( GTK_SCROLLED_WINDOW( sw ),
						GTK_SHADOW_IN );

	private->model = GTK_TREE_MODEL( gtk_tree_store_new( TREE_VIEW_MAX_COL,
							     GDK_TYPE_PIXBUF,
							     G_TYPE_STRING,
							     G_TYPE_STRING,
							     G_TYPE_UINT,
							     G_TYPE_UINT,
							     G_TYPE_BOOLEAN) );
	private->view = GTK_TREE_VIEW(
				      gtk_tree_view_new_with_model( GTK_TREE_MODEL(private->model) ) );
	gtk_tree_view_set_rules_hint( GTK_TREE_VIEW( private->view ), TRUE );

	prend = gtk_cell_renderer_pixbuf_new();
	rend = gtk_cell_renderer_text_new ();
	col = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title( col, "Element" );
	gtk_tree_view_column_pack_start( col, prend, FALSE );
	gtk_tree_view_column_pack_start( col, rend, TRUE );
	gtk_tree_view_column_set_attributes( col, rend,
					     "text", TREE_VIEW_NAME_COL,
					     NULL );
	gtk_tree_view_column_set_attributes( col, prend,
					     "pixbuf",
					     TREE_VIEW_ICON_COL,
					     NULL );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( private->view ), col );

	rend = gtk_cell_renderer_text_new ();
	col = gtk_tree_view_column_new_with_attributes( "Text",
							rend,
							"text", 
							TREE_VIEW_TEXT_COL,
							NULL );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( private->view ), col );

	rend = gtk_cell_renderer_toggle_new();
	col = gtk_tree_view_column_new_with_attributes( "Valid",
							rend,
							"active",
							TREE_VIEW_VALID_COL,
							NULL );
	gtk_tree_view_append_column( GTK_TREE_VIEW( private->view ), col );
	
	gtk_container_add( GTK_CONTAINER( sw ), GTK_WIDGET( private->view ) );
	gtk_container_add( GTK_CONTAINER( tree_view ), sw );

	gtk_widget_show( GTK_WIDGET( private->view ) );
	gtk_widget_show( sw );

	private->timeout = -1;
}

static void screem_tree_view_finalize( GObject *tree_view )
{
	ScreemTreeView *view;

	view = SCREEM_TREE_VIEW( tree_view );
	
	if( view->private->timeout != -1 ) {
		g_source_remove( view->private->timeout );
	}

	if( view->private->timeout != -1 ) {
		g_source_remove( view->private->timeout );
	}
	
	g_free( view->private );

	G_OBJECT_CLASS( parent_class )->finalize( tree_view );
}

static void screem_tree_view_window_set( ScreemView *view )
{
	/* add miniview tab */
	screem_tree_view_create_mini_view( SCREEM_TREE_VIEW( view ) );
}

static void screem_tree_view_size_request( GtkWidget *widget, 
					   GtkRequisition *req )
{
	GTK_WIDGET_CLASS( parent_class )->size_request( widget, req );
}

static void screem_tree_view_realize( GtkWidget *widget )
{
	GTK_WIDGET_CLASS( parent_class )->realize( widget );
}

static void screem_tree_view_show( GtkWidget *widget )
{
	ScreemTreeView *view;

	view = SCREEM_TREE_VIEW( widget );

	GTK_WIDGET_CLASS( parent_class )->show( widget );

	gtk_widget_set_sensitive( GTK_WIDGET( view->private->mini ), FALSE );
	
	gtk_widget_show( GTK_WIDGET( view->private->view ) );
}

static void screem_tree_view_hide( GtkWidget *widget )
{
	ScreemTreeView *view;

	view = SCREEM_TREE_VIEW( widget );

	GTK_WIDGET_CLASS( parent_class )->hide( widget );

	gtk_widget_set_sensitive( GTK_WIDGET( view->private->mini ), TRUE );
	
	gtk_widget_hide( GTK_WIDGET( view->private->view ) );
}

GType screem_tree_view_get_type()
{
	static guint type = 0;

	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemTreeViewClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_tree_view_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemTreeView ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_tree_view_init
		};
		
		type = g_type_register_static( PARENT_TYPE,
					       "ScreemTreeView",
					       &info, 0 );
	}

	return type;
}
